from dataclasses import dataclass
from typing import Iterator, Iterable, Generic, TypeVar
from sqlschm import tok

T = TypeVar("T")


@dataclass(slots=True)
class ItemCursor(Generic[T]):
    _it: Iterator[T]
    _default_item: T
    item: T
    # Enable to look ahead
    next_item: T

    def __init__(self, items: Iterable[T], default_item: T):
        self._it = iter(items)
        self._default_item = default_item
        self.next_item = default_item
        self.forth()
        self.forth()

    def forth(self) -> None:
        self.item = self.next_item
        self.next_item = next(self._it, self._default_item)


def tokens(src: Iterable[str], /) -> Iterable[tok.Token]:
    cs = ItemCursor(src, None)
    while cs.item is not None:
        assert len(cs.item) == 1
        if cs.item.isspace():
            yield tok.INTERNED[cs.item]
            cs.forth()
        elif cs.item.isdecimal() or (
            cs.item == "." and cs.next_item is not None and cs.next_item.isdecimal()
        ):
            yield _numeric_token(cs)
        elif cs.item.isidentifier():
            id = _identifier(cs)
            id_upper = id.upper()
            if id_upper == "B" and cs.item in ["'", '"']:
                yield _bin_token(cs)
            elif id_upper == "X" and cs.item in ["'", '"']:
                yield _blob_token(cs)
            elif id_upper in tok.INTERNED:
                yield tok.INTERNED[id_upper]
            else:
                yield tok.Token(tok.TokenKind.RAW_ID, id)
        elif cs.item in "'\"":
            yield _str_token(cs)
        elif cs.item in "`[":
            yield _enclosed_id_token(cs)
        elif cs.item in tok.INTERNED:
            concat = cs.item + cs.next_item if cs.next_item is not None else cs.item
            if concat == "--" or concat == "# ":
                yield _single_line_comment_token(cs)
            elif concat == "/*":
                yield _multi_line_comment_token(cs)
            else:
                if concat in tok.INTERNED:
                    yield tok.INTERNED[concat]
                    cs.forth()
                else:
                    yield tok.INTERNED[cs.item]
                cs.forth()
        else:
            yield tok.Token(tok.TokenKind.UNKNOWN, cs.item)
            cs.forth()


def _single_line_comment_token(cs: ItemCursor[str | None], /) -> tok.Token:
    cs.forth()  # consume -
    cs.forth()  # consume -
    content = ""
    while cs.item is not None and cs.item != "\n":
        content += cs.item
        cs.forth()
    cs.forth()  # consume  \n or None
    return tok.Token(tok.TokenKind.SINGLE_LINE_COMMENT, content)


def _multi_line_comment_token(cs: ItemCursor[str | None], /) -> tok.Token:
    cs.forth()  # consume /
    cs.forth()  # consume *
    content = ""
    while (
        cs.item is not None
        and cs.next_item is not None
        and (cs.item + cs.next_item) != "*/"
    ):
        content += cs.item
        cs.forth()
    if cs.item is None or cs.next_item is None:
        if cs.item is not None:
            content += cs.item
            cs.forth()
        return tok.Token(tok.TokenKind.UNKNOWN, f"/*{content}")
    cs.forth()  # consume *
    cs.forth()  # consume /
    return tok.Token(tok.TokenKind.MULTI_LINE_COMMENT, content)


def _blob_token(cs: ItemCursor[str | None], /) -> tok.Token:
    assert cs.item in ["'", '"']
    delim = cs.item
    cs.forth()  # consume ' or "
    n = _hex_literal(cs)
    if cs.item != delim:
        return tok.Token(tok.TokenKind.UNKNOWN, f"{delim}{n}")
    cs.forth()  # consume ' or "
    return tok.Token(tok.TokenKind.BLOB, n)


def _bin_token(cs: ItemCursor[str | None], /) -> tok.Token:
    assert cs.item in ["'", '"']
    delim = cs.item
    cs.forth()  # consume ' or "
    n = _bin_literal(cs)
    if cs.item != delim:
        return tok.Token(tok.TokenKind.UNKNOWN, f"{delim}{n}")
    cs.forth()  # consume ' or "
    return tok.Token(tok.TokenKind.BINARY, n)


def _numeric_token(cs: ItemCursor[str | None], /) -> tok.Token:
    first_part = _fractional_literal(cs)
    if "e" in first_part:
        return tok.Token(tok.TokenKind.FLOAT, first_part)
    if cs.item == ".":
        cs.forth()
        val = first_part + "." + _fractional_literal(cs)
        return tok.Token(tok.TokenKind.FLOAT, val)
    elif first_part == "0" and cs.item in ["x", "X"]:
        x = cs.item
        cs.forth()
        if cs.item in _HEX_DIGITS:
            return tok.Token(tok.TokenKind.HEX, _hex_literal(cs))
        else:
            return tok.Token(tok.TokenKind.UNKNOWN, f"0{x}")
    return tok.Token(tok.TokenKind.INT, first_part)


def _str_token(cs: ItemCursor[str | None], /) -> tok.Token:
    assert cs.item is not None
    delim = cs.item  # " or '
    cs.forth()  # consume delim
    s = ""
    while cs.item is not None and (cs.item != delim or cs.next_item == delim):
        # escaped delim?
        if cs.item == delim:
            assert cs.next_item == delim
            cs.forth()
        s += cs.item
        cs.forth()
    if cs.item is None:
        return tok.Token(tok.TokenKind.UNKNOWN, f"{delim}{s}")
    cs.forth()  # consume delim
    kind = tok.TokenKind.STD_STR if delim == "'" else tok.TokenKind.STD_DELIMITED_ID
    return tok.Token(kind, s)


def _enclosed_id_token(cs: ItemCursor[str | None], /) -> tok.Token:
    delim = cs.item
    cs.forth()
    s = _identifier(cs)
    if cs.item != delim:
        pass
    cs.forth()
    return tok.Token(tok.TokenKind.NON_STD_DELIMITED_ID, s)


def _identifier(cs: ItemCursor[str | None], /) -> str:
    result: str = ""
    while cs.item is not None and (cs.item in "_$" or cs.item.isalnum()):
        result += cs.item
        cs.forth()
    return result


def _fractional_literal(cs: ItemCursor[str | None], /) -> str:
    decimal = _int_literal(cs)
    if cs.item in ["e", "E"]:
        cs.forth()
        sign = ""
        if cs.item in ["+", "-"]:
            sign = cs.item
            cs.forth()
        exponent = _int_literal(cs)
        return decimal + "e" + sign + exponent
    return decimal


def _hex_literal(cs: ItemCursor[str | None], /) -> str:
    result: str = ""
    while cs.item in _HEX_DIGITS:
        result += cs.item
        cs.forth()
    return result


def _bin_literal(cs: ItemCursor[str | None], /) -> str:
    result: str = ""
    while cs.item in ["0", "1"]:
        result += cs.item
        cs.forth()
    return result


def _int_literal(cs: ItemCursor[str | None], /) -> str:
    result: str = ""
    while cs.item is not None and cs.item.isdecimal():
        result += cs.item
        cs.forth()
    return result


_HEX_DIGITS: frozenset[str] = frozenset(iter("0123456789ABCDEFabcdef"))



from collections.abc import Sequence
import textwrap
from sqlschm import sql, tok


def generate_schema(schema: sql.Schema, dialect: sql.Dialect, /) -> str:
    result = ""
    for table in schema.tables:
        result += _generate_create_table(table)
    return result.strip()


def _generate_create_table(table: sql.Table, /) -> str:
    table_name = _generate_qualified_name(table.name)
    temp = " TEMPORARY" if table.temporary else ""
    if_not_exists = " IF NOT EXISTS" if table.if_not_exists else ""
    or_replace = " OR REPLACE" if table.or_replace else ""
    option_list: list[str] = []
    if table.options.strict:
        option_list += [" STRICT"]
    if table.options.without_rowid:
        option_list += [" WITHOUT ROWID"]
    options = ",".join(option_list)
    body_entries = list(map(_generate_column_def, table.columns)) + list(
        map(_generate_table_constraint, table.constraints)
    )
    body = ",\n".join(body_entries)
    body = "\n" + textwrap.indent(body, " " * 12)
    return textwrap.dedent(
        f"""
        CREATE{or_replace}{temp} TABLE{if_not_exists} {table_name}({body}
        ){options};
        """
    )


def _generate_column_def(col: sql.Column, /) -> str:
    coltype = " " + _generate_type(col.type) if col.type.name != "" else ""
    constraints = "".join(_generate_column_constraint(x) for x in col.constraints)
    return f'"{col.name}"{coltype}{constraints}'


def _generate_column_constraint(constraint: sql.ColumnConstraint, /) -> str:
    name = f' CONSTRAINT"{constraint.name}" ' if constraint.name is not None else ""
    if isinstance(constraint, sql.Uniqueness):
        on_conflict = _generate_on_conflict(constraint.on_conflict)
        if constraint.is_primary:
            sorting = constraint.indexed[0].sorting
            sorting_str = f"{sorting.name}" if sorting is not None else ""
            autoinc = " AUTOINCREMENT" if constraint.autoincrement else ""
            return f"{name} PRIMARY KEY{sorting_str}{on_conflict}{autoinc}"
        else:
            return f"{name} UNIQUE{on_conflict}"
    elif isinstance(constraint, sql.ForeignKey):
        fk_clause = _generate_foreign_key_clause(constraint)
        return f"{name} {fk_clause}"
    elif isinstance(constraint, sql.Check):
        expr = _generate_expr(constraint.expr)
        return f"{name} CHECK ({expr})"
    elif isinstance(constraint, sql.NotNull):
        on_conflict = _generate_on_conflict(constraint.on_conflict)
        return f"{name} NOT NULL{on_conflict}"
    elif isinstance(constraint, sql.Default):
        expr = _generate_expr(constraint.expr)
        return f"{name} DEFAULT {expr}"
    elif isinstance(constraint, sql.Collation):
        return f"{name} COLLATE {constraint.value}"
    else:
        kind = f" {constraint.kind.name}" if constraint.kind is not None else ""
        expr = _generate_expr(constraint.expr)
        return f"{name} GENERATED ALWAYS AS ({expr}){kind}"


def _generate_table_constraint(constraint: sql.TableConstraint, /) -> str:
    name = f'CONSTRAINT "{constraint.name}" ' if constraint.name is not None else ""
    if isinstance(constraint, sql.Uniqueness):
        idxs = ", ".join(_generate_indexed(idx) for idx in constraint.indexed)
        on_conflict = _generate_on_conflict(constraint.on_conflict)
        if constraint.is_primary:
            return f"{name}PRIMARY KEY ({idxs}){on_conflict}"
        else:
            return f"{name}UNIQUE ({idxs}){on_conflict}"
    elif isinstance(constraint, sql.ForeignKey):
        cols = ", ".join(f'"{col}"' for col in constraint.columns)
        return f"{name}FOREIGN KEY ({cols}) " + _generate_foreign_key_clause(constraint)
    else:
        expr = _generate_expr(constraint.expr)
        return f"{name}CHECK ({expr})"


def _generate_foreign_key_clause(constraint: sql.ForeignKey, /) -> str:
    foreign_table = _generate_qualified_name(constraint.foreign_table.name)
    referred_columns = (
        ("(" + ", ".join(f'"{col}"' for col in constraint.referred_columns) + ")")
        if constraint.referred_columns is not None
        else ""
    )
    on_update = _generate_on_update_delete(constraint.on_update, True)
    on_delete = _generate_on_update_delete(constraint.on_delete, False)
    match = f" MATCH {constraint.match.name}" if constraint.match is not None else ""
    enforcement = _generate_constraint_enforcement(constraint.enforcement)
    return (
        f"REFERENCES {foreign_table}{referred_columns}"
        + f"{on_update}{on_delete}{match}{enforcement}"
    )


def _generate_constraint_enforcement(
    enforcement: sql.ConstraintEnforcement | None,
) -> str:
    if enforcement is not None:
        not_deferrable = (
            " NOT DEFERRABLE" if enforcement.not_deferrable else " DEFERRABLE"
        )
        if enforcement.initially is not None:
            initially = f" INITIALLY {enforcement.initially.name}"
        else:
            initially = ""
        return f"{not_deferrable}{initially}"
    return ""


def _generate_indexed(indexed: sql.Indexed, /) -> str:
    collation = ""
    sorting = ""
    if indexed.collation is not None:
        collation = f" COLLATE {indexed.collation.name}"
    if indexed.sorting is not None:
        sorting = f" {indexed.sorting.name}"
    return f'"{indexed.column}"{collation}{sorting}'


def _generate_on_conflict(on_conflict: sql.OnConflict | None, /) -> str:
    if on_conflict is None:
        return ""
    else:
        return f" ON CONFLICT {on_conflict.name}"


def _generate_on_update_delete(
    action: sql.OnUpdateDelete | None, on_update: bool, /
) -> str:
    if action is not None:
        action_name = action.name.replace("_", " ")
        if on_update:
            return f" ON UPDATE {action_name}"
        else:
            return f" ON DELETE {action_name}"
    return ""


def _generate_type(type: sql.Type | None, /) -> str:
    if type is not None:
        iter(type.params)
        params = (
            f"({', '.join(map(str, type.params))})" if len(type.params) != 0 else ""
        )
        return type.name.lower() + params
    return ""


def _generate_qualified_name(qualified_name: sql.QualifiedName, /) -> str:
    names = list(qualified_name)
    names.reverse()
    return ".".join(f'"{name}"' for name in names)


def _generate_expr(expr: Sequence[tok.Token], /) -> str:
    return " ".join(_generate_tok(tk) for tk in expr)


def _generate_tok(tk: tok.Token, /) -> str:
    if bool(tk.kind & tok.TokenKind.DELIMITED_ID):
        return f'"{tk.val}"'
    elif bool(tk.kind & tok.TokenKind.STR):
        return f"'{tk.val}'"
    elif bool(tk.kind & tok.TokenKind.BLOB):
        return f"X'{tk.val}'"
    elif bool(tk.kind & tok.TokenKind.BINARY):
        return f"B'{tk.val}'"
    elif bool(tk.kind & tok.TokenKind.HEX):
        return f"0x{tk.val}"
    elif bool(tk.kind & tok.TokenKind.NUMERIC):
        return f"{tk.val}"
    elif bool(tk.kind & tok.TokenKind.TRIVIA):
        return ""
    else:
        return tk.val



from typing import Iterable
from sqlschm import sql, tok, lexer

Lex = lexer.ItemCursor[tok.Token]


class ParserError(Exception):
    pass


# sentinel to mark the end of a token stream
_EOF_TOKEN: tok.Token = tok.Token(tok.TokenKind.UNKNOWN, "")


def parse_schema(src: Iterable[str], /) -> sql.Schema:
    non_trivia_tokens = filter(tok.is_not_trivia, lexer.tokens(src))
    l = lexer.ItemCursor(non_trivia_tokens, _EOF_TOKEN)
    tables: list[sql.Table] = []
    while l.item is not _EOF_TOKEN:
        tables.append(_parse_create_table(l))
    return sql.Schema(tables=tables)


def _parse_create_table(l: Lex, /) -> sql.Table:
    if_not_exists = False
    or_replace = False
    temporary = False
    _expect(l, tok.CREATE)
    if l.item is tok.OR:
        l.forth()
        _expect(l, tok.REPLACE)
        or_replace = True
    if l.item is tok.TEMPORARY or l.item is tok.TEMP:
        l.forth()
        temporary = True
    _expect(l, tok.TABLE)
    if l.item is tok.IF:
        l.forth()
        _expect(l, tok.NOT)
        _expect(l, tok.EXISTS)
        if_not_exists = True
    table_name = _parse_qualified_name(l)
    if l.item is tok.AS or l.item is tok.LIKE:
        # FIXME: support this case?
        raise ParserError(f"'CREATE TABLE {l.item.val}' is not supported.")
    _expect(l, tok.L_PAREN)
    columns: list[sql.Column] = []
    constraints: list[sql.TableConstraint] = []
    if l.item is not tok.R_PAREN:
        columns.append(_parse_column_def(l))
        while l.item is tok.COMMA and l.next_item.kind is not tok.TokenKind.KEYWORD:
            l.forth()
            columns.append(_parse_column_def(l))
        while l.item is tok.COMMA:
            l.forth()
            constraints.append(_parse_table_constraint(l))
        _expect(l, tok.R_PAREN)
    options = _parse_table_options(l)
    if l.item is tok.SELECT:
        # consume SELECT expression
        while l.item is not tok.SEMICOLON:
            l.forth()
    _expect(l, tok.SEMICOLON)
    return sql.Table(
        name=table_name,
        if_not_exists=if_not_exists,
        or_replace=or_replace,
        temporary=temporary,
        columns=columns,
        constraints=constraints,
        options=options,
    )


def _parse_column_def(l: Lex, /) -> sql.Column:
    colname = _parse_name(l)
    coltype = _parse_type(l)
    constraints: list[sql.ColumnConstraint] = []
    while l.item is not tok.COMMA and l.item is not tok.R_PAREN:
        constraints.append(_parse_col_constraint(l, colname))
    return sql.Column(name=colname, type=coltype, constraints=constraints)


def _parse_table_options(l: Lex, /) -> sql.TableOptions:
    strict = False
    without_rowid = False
    while True:
        if l.item is tok.STRICT:
            l.forth()
            strict = True
        elif l.item is tok.WITHOUT:
            l.forth()
            _expect(l, tok.ROWID)
            without_rowid = True
        if l.item is tok.COMMA:
            l.forth()
        else:
            break
    return sql.TableOptions(strict=strict, without_rowid=without_rowid)


def _parse_type(l: Lex, /) -> sql.Type:
    # see https://www.sqlite.org/datatype3.html
    # TODO: while it is not a keyword, a comma, or a rparen -> it is the type!
    type_name = ""
    type_params: list[int] = []
    while bool(l.item.kind & tok.TokenKind.NON_KW_ID):
        type_name += l.item.val.upper() + " "
        l.forth()
    type_name = type_name.rstrip()
    if type_name != "":
        if l.item is tok.L_PAREN:
            l.forth()
            type_params.append(_parse_int(l))
            if l.item is tok.COMMA:
                type_params.append(_parse_int(l))
            _expect(l, tok.R_PAREN)
    return sql.Type(name=type_name, params=type_params)


def _parse_table_constraint(l: Lex, /) -> sql.TableConstraint:
    name = None
    if l.item is tok.CONSTRAINT:
        l.forth()
        name = _parse_name(l)
    if l.item is tok.PRIMARY or l.item is tok.UNIQUE:
        is_primary = l.item is tok.PRIMARY
        l.forth()
        if is_primary:
            _expect(l, tok.KEY)
        return sql.Uniqueness(
            name=name,
            is_table_constraint=True,
            indexed=_parse_indexed_names(l),
            is_primary=is_primary,
            on_conflict=_parse_on_conflict(l),
        )
    elif l.item is tok.FOREIGN:
        l.forth()
        _expect(l, tok.KEY)
        columns = _parse_parens_names(l)
        return _parse_foreign_key_clause(l, columns, name, is_table_constraint=True)
    elif l.item is tok.CHECK:
        l.forth()
        expr = tokens_in_parens(l)
        return sql.Check(name=name, is_table_constraint=True, expr=expr)
    else:
        raise ParserError(f"'{l.item.val}' cannot start a table constraint")


def _parse_col_constraint(l: Lex, col_name: str, /) -> sql.ColumnConstraint:
    columns = [col_name]
    name = None
    if l.item is tok.CONSTRAINT:
        l.forth()
        name = _parse_name(l)
    if l.item is tok.PRIMARY or l.item is tok.UNIQUE:
        is_primary = l.item is tok.PRIMARY
        l.forth()
        if is_primary:
            _expect(l, tok.KEY)
        sorting = _parse_optional_sorting(l)
        indexed = [sql.Indexed(column=col_name, sorting=sorting)]
        on_conflict = _parse_on_conflict(l)
        autoincrement = l.item is tok.AUTOINCREMENT or l.item is tok.AUTO_INCREMENT
        if autoincrement:
            l.forth()
        return sql.Uniqueness(
            name=name,
            indexed=indexed,
            is_primary=is_primary,
            autoincrement=autoincrement,
            on_conflict=on_conflict,
        )
    elif l.item is tok.CHECK:
        l.forth()
        expr = tokens_in_parens(l)
        return sql.Check(name=name, expr=expr)
    elif l.item is tok.REFERENCES:
        return _parse_foreign_key_clause(l, columns, name)
    elif l.item is tok.NOT:
        l.forth()
        _expect(l, tok.NULL)
        return sql.NotNull(name=name, on_conflict=_parse_on_conflict(l))
    elif l.item is tok.DEFAULT:
        l.forth()
        return sql.Default(name=name, expr=_parse_expr(l))
    elif l.item is tok.COLLATE:
        l.forth()
        return sql.Collation(name=name, value=_parse_name(l))
    elif l.item is tok.GENERATED or l.item is tok.AS:
        if l.item is tok.GENERATED and l.next_item is tok.ALWAYS:
            l.forth()
            l.forth()
        elif l.item is tok.GENERATED and l.next_item is tok.BY:
            l.forth()
            l.forth()
            _expect(l, tok.DEFAULT)
        _expect(l, tok.AS)
        if l.item is tok.IDENTITY:
            l.forth()
        expr = tokens_in_parens(l)
        kind = None
        tok_val = l.item.val.upper()
        if bool(l.item.kind & tok.TokenKind.ID) and tok_val in sql.GENERATED_KIND:
            kind = sql.GeneratedKind[tok_val]
            l.forth()
        return sql.Generated(name=name, expr=expr, kind=kind)
    else:
        raise ParserError(f"'{l.item.val}' cannot start a column constraint")


def _parse_indexed_names(l: Lex, /) -> list[sql.Indexed]:
    _expect(l, tok.L_PAREN)
    result = [_parse_indexed_name(l)]
    while l.item is tok.COMMA:
        l.forth()
        result.append(_parse_indexed_name(l))
    _expect(l, tok.R_PAREN)
    return result


def _parse_indexed_name(l: Lex, /) -> sql.Indexed:
    column = _parse_name(l)
    collation = None
    if l.item is tok.COLLATE:
        l.forth()
        collation = sql.Collation(value=_parse_name(l))
    sorting = _parse_optional_sorting(l)
    return sql.Indexed(column=column, collation=collation, sorting=sorting)


def _parse_optional_sorting(l: Lex, /) -> sql.Sorting | None:
    if l.item is tok.ASC:
        l.forth()
        return sql.Sorting.ASC
    elif l.item is tok.DESC:
        l.forth()
        return sql.Sorting.DESC
    return None


def _parse_expr(l: Lex, /) -> list[tok.Token]:
    result: list[tok.Token] = []
    if bool(l.item.kind & tok.TokenKind.LITERAL):
        result.append(l.item)
        l.forth()
    elif l.item is tok.NUM_PLUS or l.item is tok.NUM_MINUS:
        result += [l.item, l.next_item]
        l.forth()
        _parse_int(l)  # ensure it is an integer
    elif bool(l.item.kind & tok.TokenKind.ID) and l.next_item is tok.L_PAREN:
        # function call
        result.append(l.item)
        l.forth()
        result.append(tok.L_PAREN)
        result += tokens_in_parens(l)
        result.append(tok.R_PAREN)
    elif l.item is tok.L_PAREN:
        result += tokens_in_parens(l)
    else:
        raise ParserError(f"'{l.item.val}' is not a supported DEFAULT value.")
    return result


def _parse_parens_names(l: Lex, /) -> list[str]:
    _expect(l, tok.L_PAREN)
    result = [_parse_name(l)]
    while l.item is tok.COMMA:
        l.forth()
        result.append(_parse_name(l))
    _expect(l, tok.R_PAREN)
    return result


def _parse_foreign_key_clause(
    l: Lex,
    columns: list[str],
    name: str | None,
    /,
    *,
    is_table_constraint: bool = False,
) -> sql.ForeignKey:
    _expect(l, tok.REFERENCES)
    foreign_table = sql.Alias(name=_parse_qualified_name(l))
    referred_columns: list[str] | None = None
    on_delete: sql.OnUpdateDelete | None = None
    on_update: sql.OnUpdateDelete | None = None
    match: sql.Match | None = None
    if l.item is tok.L_PAREN:
        referred_columns = _parse_parens_names(l)
    while l.item is tok.ON or l.item is tok.MATCH:
        if l.item is tok.ON:
            l.forth()
            if l.item is tok.DELETE:
                l.forth()
                on_delete = _parse_on_updatedelete_action(l)
            elif l.item is tok.UPDATE:
                l.forth()
                on_update = _parse_on_updatedelete_action(l)
            else:
                raise ParserError("'ON DELETE' or 'ON UPDATE' is expected")
        elif l.item is tok.MATCH:
            l.forth()
            match_name = _parse_name(l).upper()
            if match_name not in sql.MATCH:
                raise ParserError("invalid MATCH name")
            match = sql.Match[match_name]
    enforcement = _parse_constraint_enforcement(l)
    return sql.ForeignKey(
        name=name,
        columns=columns,
        foreign_table=foreign_table,
        referred_columns=referred_columns,
        is_table_constraint=is_table_constraint,
        on_delete=on_delete,
        on_update=on_update,
        match=match,
        enforcement=enforcement,
    )


def _parse_constraint_enforcement(l: Lex, /) -> sql.ConstraintEnforcement | None:
    initially = _parse_constraint_enforcement_time(l)
    not_deferrable = None
    if l.item is tok.NOT and l.next_item is tok.DEFERRABLE:
        l.forth()
        not_deferrable = True
    if l.item is tok.DEFERRABLE:
        l.forth()
        not_deferrable = False
        if initially is None:
            initially = _parse_constraint_enforcement_time(l)
    if not_deferrable is not None:
        return sql.ConstraintEnforcement(
            initially=initially, not_deferrable=not_deferrable
        )
    return None


def _parse_constraint_enforcement_time(
    l: Lex, /
) -> sql.ConstraintEnforcementTime | None:
    if l.item is tok.INITIALLY:
        l.forth()
        if l.item is tok.DEFERRED:
            l.forth()
            return sql.ConstraintEnforcementTime.DEFERRED
        elif l.item is tok.IMMEDIATE:
            l.forth()
            return sql.ConstraintEnforcementTime.IMMEDIATE
        else:
            raise ParserError(f"'{l.item.val}' is not a valid enforcement time")
    return None


def _parse_on_updatedelete_action(l: Lex, /) -> sql.OnUpdateDelete:
    if l.item is tok.CASCADE:
        l.forth()
        return sql.OnUpdateDelete.CASCADE
    elif l.item is tok.NO and l.next_item is tok.ACTION:
        l.forth()
        l.forth()
        return sql.OnUpdateDelete.NO_ACTION
    elif l.item is tok.SET and l.next_item is tok.NULL:
        l.forth()
        l.forth()
        return sql.OnUpdateDelete.SET_NULL
    elif l.item is tok.SET and l.next_item is tok.DEFAULT:
        l.forth()
        l.forth()
        return sql.OnUpdateDelete.SET_DEFAULT
    elif l.item is tok.RESTRICT:
        l.forth()
        return sql.OnUpdateDelete.RESTRICT
    else:
        raise ParserError("'Invalid ON DELETE/UPDATE action")


def _parse_on_conflict(l: Lex, /) -> sql.OnConflict | None:
    if l.item is tok.ON:
        l.forth()
        _expect(l, tok.CONFLICT)
        if l.item.val not in sql.ON_CONFLICT:
            raise ParserError(f"'{l.item.val}' is not a valid ON CONFLICT ACTION")
        result = sql.OnConflict[l.item.val]
        l.forth()
        return result
    else:
        return None


def _parse_int(l: Lex, /) -> int:
    if l.item is not tok.TokenKind.INT:
        raise ParserError("an integer is expected.")
    result = int(l.item)
    l.forth()
    return result


def _parse_qualified_name(l: Lex, /) -> sql.QualifiedName:
    names = [_parse_name(l)]
    while l.item is tok.DOT:
        l.forth()
        names.append(_parse_name(l))
    names.reverse()
    return names


def _parse_name(l: Lex, /) -> str:
    if not bool(l.item.kind & tok.TokenKind.ID):
        raise ParserError(f"an identifier is expected.")
    result = l.item.val
    l.forth()
    return result


def skip_expr(l: Lex, /) -> None:
    if l.item is tok.L_PAREN:
        skip_parens(l)
    else:
        if l.item is tok.INTERNED["+"] or l.item is tok.INTERNED["-"]:
            l.forth()
        _skip(l, tok.TokenKind.LITERAL)


def skip_parens(l: Lex, /) -> None:
    _expect(l, tok.L_PAREN)
    count = 0
    while l.item is not tok.R_PAREN or count > 0:
        if l.item is tok.L_PAREN:
            count += 1
        elif l.item is tok.R_PAREN:
            count -= 1
        l.forth()
    l.forth()


def tokens_in_parens(l: Lex, /) -> list[tok.Token]:
    result: list[tok.Token] = []
    _expect(l, tok.L_PAREN)
    count = 0
    while l.item is not tok.R_PAREN or count > 0:
        result.append(l.item)
        if l.item is tok.L_PAREN:
            count += 1
        elif l.item is tok.R_PAREN:
            count -= 1
        l.forth()
    l.forth()
    return result


def _expect(l: Lex, tk: tok.Token, /) -> None:
    if l.item is not tk:
        raise ParserError(f"'{tk.val}' is expected. Got '{l.item.val}'.")
    l.forth()


def _skip(l: Lex, kind: tok.TokenKind, /) -> None:
    if l.item.kind is kind:
        raise ParserError(f"a {kind.name} is expected")
    l.forth()



from dataclasses import dataclass
from enum import Flag, unique, auto


class _ReprFlag(Flag):
    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


@unique
class TokenKind(_ReprFlag):
    KEYWORD = auto()

    RAW_ID = auto()  # id
    STD_DELIMITED_ID = auto()  # "id"

    NON_STD_DELIMITED_ID = auto()  # `id` or [id]
    DELIMITED_ID = STD_DELIMITED_ID | NON_STD_DELIMITED_ID

    STD_STR = auto()  # 'a string'
    STR = STD_STR | STD_DELIMITED_ID  # 'a string' or "a string"

    NON_KW_ID = RAW_ID | DELIMITED_ID | STD_STR
    ID = NON_KW_ID | KEYWORD

    PREDEF_LITERAL = auto()
    KEYWORD_LITERAL = KEYWORD | PREDEF_LITERAL  # NULL, CURRENT_*
    RAW_ID_LITERAL = RAW_ID | PREDEF_LITERAL  # TRUE, FALSE

    BLOB = auto()
    BINARY = auto()
    FLOAT = auto()  # .5e+6
    HEX = auto()  # 0x5A
    INT = auto()  # 123
    NUMERIC = FLOAT | HEX | INT
    LITERAL = BLOB | BINARY | NUMERIC | STR | PREDEF_LITERAL

    BIN_OP = auto()
    CMP_OP = auto()
    NUM_OP = auto()
    STR_OP = auto()
    OPERATOR = BIN_OP | CMP_OP | NUM_OP | STR_OP

    PUNCTUATION = auto()

    SINGLE_LINE_COMMENT = auto()  # -- or #<space>
    MULTI_LINE_COMMENT = auto()  # /* */
    COMMENT = SINGLE_LINE_COMMENT | MULTI_LINE_COMMENT

    WHITESPACE = auto()
    TRIVIA = WHITESPACE | COMMENT

    UNKNOWN = auto()
    UNKNOWN_OR_TRIVIA = UNKNOWN | TRIVIA


@dataclass(frozen=True, slots=True)
class Token:
    kind: TokenKind
    val: str


def is_not_trivia(tk: Token, /) -> bool:
    return not bool(tk.kind & TokenKind.TRIVIA)


def like(a: Token, b: Token, /) -> bool:
    return bool(a.kind & b.kind) and a.val == b.val


# Spaces
NEWLINE: Token = Token(TokenKind.WHITESPACE, "\n")

# *
WILDCARD: Token = Token(TokenKind.NUM_OP, "*")

# Binary operators
BIN_NEG: Token = Token(TokenKind.BIN_OP, "~")
BIN_AND: Token = Token(TokenKind.BIN_OP, "&")
BIN_OR: Token = Token(TokenKind.BIN_OP, "|")
BIN_LSHIFT: Token = Token(TokenKind.BIN_OP, "<<")
BIN_RSHIFT: Token = Token(TokenKind.BIN_OP, ">>")

# Numeric operators
NUM_PLUS: Token = Token(TokenKind.NUM_OP, "+")
NUM_MINUS: Token = Token(TokenKind.NUM_OP, "-")
NUM_DIV: Token = Token(TokenKind.NUM_OP, "/")
NUM_MOD: Token = Token(TokenKind.NUM_OP, "%")
NUM_EXP: Token = Token(TokenKind.NUM_OP, "^")

# Comparison operators
CMP_EQ: Token = Token(TokenKind.CMP_OP, "=")
CMP_EQ2: Token = Token(TokenKind.CMP_OP, "==")
CMP_GEQ: Token = Token(TokenKind.CMP_OP, ">=")
CMP_GEQ2: Token = Token(TokenKind.CMP_OP, "!>")
CMP_GT: Token = Token(TokenKind.CMP_OP, ">")
CMP_LEQ: Token = Token(TokenKind.CMP_OP, "<=")
CMP_LEQ2: Token = Token(TokenKind.CMP_OP, "!<")
CMP_LT: Token = Token(TokenKind.CMP_OP, "<")
CMP_NEQ: Token = Token(TokenKind.CMP_OP, "<>")
CMP_NEQ2: Token = Token(TokenKind.CMP_OP, "!=")

# String operators
STR_CONCAT: Token = Token(TokenKind.STR_OP, "||")

# Punctuation
DOT: Token = Token(TokenKind.PUNCTUATION, ".")
COMMA: Token = Token(TokenKind.PUNCTUATION, ",")
COLON: Token = Token(TokenKind.PUNCTUATION, ":")
DOUBLE_COLON: Token = Token(TokenKind.PUNCTUATION, "::")
SEMICOLON: Token = Token(TokenKind.PUNCTUATION, ";")
L_PAREN: Token = Token(TokenKind.PUNCTUATION, "(")
R_PAREN: Token = Token(TokenKind.PUNCTUATION, ")")

# Known raw identifiers
ROWID: Token = Token(TokenKind.RAW_ID, "ROWID")
STRICT: Token = Token(TokenKind.RAW_ID, "STRICT")
FALSE: Token = Token(TokenKind.RAW_ID_LITERAL, "FALSE")
TRUE: Token = Token(TokenKind.RAW_ID_LITERAL, "TRUE")
IDENTITY: Token = Token(TokenKind.RAW_ID, "IDENTITY")

# Keywords
ABORT: Token = Token(TokenKind.KEYWORD, "ABORT")
ACTION: Token = Token(TokenKind.KEYWORD, "ACTION")
ADD: Token = Token(TokenKind.KEYWORD, "ADD")
AFTER: Token = Token(TokenKind.KEYWORD, "AFTER")
ALL: Token = Token(TokenKind.KEYWORD, "ALL")
ALTER: Token = Token(TokenKind.KEYWORD, "ALTER")
ALWAYS: Token = Token(TokenKind.KEYWORD, "ALWAYS")
ANALYZE: Token = Token(TokenKind.KEYWORD, "ANALYZE")
AND: Token = Token(TokenKind.KEYWORD, "AND")
AS: Token = Token(TokenKind.KEYWORD, "AS")
ASC: Token = Token(TokenKind.KEYWORD, "ASC")
ATTACH: Token = Token(TokenKind.KEYWORD, "ATTACH")
AUTOINCREMENT: Token = Token(TokenKind.KEYWORD, "AUTOINCREMENT")
AUTO_INCREMENT: Token = Token(TokenKind.KEYWORD, "AUTO_INCREMENT")
BEFORE: Token = Token(TokenKind.KEYWORD, "BEFORE")
BEGIN: Token = Token(TokenKind.KEYWORD, "BEGIN")
BETWEEN: Token = Token(TokenKind.KEYWORD, "BETWEEN")
BY: Token = Token(TokenKind.KEYWORD, "BY")
CASCADE: Token = Token(TokenKind.KEYWORD, "CASCADE")
CASE: Token = Token(TokenKind.KEYWORD, "CASE")
CAST: Token = Token(TokenKind.KEYWORD, "CAST")
CHECK: Token = Token(TokenKind.KEYWORD, "CHECK")
COLLATE: Token = Token(TokenKind.KEYWORD, "COLLATE")
COLUMN: Token = Token(TokenKind.KEYWORD, "COLUMN")
COMMIT: Token = Token(TokenKind.KEYWORD, "COMMIT")
CONFLICT: Token = Token(TokenKind.KEYWORD, "CONFLICT")
CONSTRAINT: Token = Token(TokenKind.KEYWORD, "CONSTRAINT")
CREATE: Token = Token(TokenKind.KEYWORD, "CREATE")
CROSS: Token = Token(TokenKind.KEYWORD, "CROSS")
CURRENT: Token = Token(TokenKind.KEYWORD, "CURRENT")
CURRENT_DATE: Token = Token(TokenKind.KEYWORD_LITERAL, "CURRENT_DATE")
CURRENT_TIME: Token = Token(TokenKind.KEYWORD_LITERAL, "CURRENT_TIME")
CURRENT_TIMESTAMP: Token = Token(TokenKind.KEYWORD_LITERAL, "CURRENT_TIMESTAMP")
DATABASE: Token = Token(TokenKind.KEYWORD, "DATABASE")
DEFAULT: Token = Token(TokenKind.KEYWORD, "DEFAULT")
DEFERRABLE: Token = Token(TokenKind.KEYWORD, "DEFERRABLE")
DEFERRED: Token = Token(TokenKind.KEYWORD, "DEFERRED")
DELETE: Token = Token(TokenKind.KEYWORD, "DELETE")
DESC: Token = Token(TokenKind.KEYWORD, "DESC")
DETACH: Token = Token(TokenKind.KEYWORD, "DETACH")
DISTINCT: Token = Token(TokenKind.KEYWORD, "DISTINCT")
DO: Token = Token(TokenKind.KEYWORD, "DO")
DROP: Token = Token(TokenKind.KEYWORD, "DROP")
EACH: Token = Token(TokenKind.KEYWORD, "EACH")
ELSE: Token = Token(TokenKind.KEYWORD, "ELSE")
END: Token = Token(TokenKind.KEYWORD, "END")
ESCAPE: Token = Token(TokenKind.KEYWORD, "ESCAPE")
EXCEPT: Token = Token(TokenKind.KEYWORD, "EXCEPT")
EXCLUDE: Token = Token(TokenKind.KEYWORD, "EXCLUDE")
EXCLUSIVE: Token = Token(TokenKind.KEYWORD, "EXCLUSIVE")
EXISTS: Token = Token(TokenKind.KEYWORD, "EXISTS")
EXPLAIN: Token = Token(TokenKind.KEYWORD, "EXPLAIN")
FAIL: Token = Token(TokenKind.KEYWORD, "FAIL")
FILTER: Token = Token(TokenKind.KEYWORD, "FILTER")
FIRST: Token = Token(TokenKind.KEYWORD, "FIRST")
FOLLOWING: Token = Token(TokenKind.KEYWORD, "FOLLOWING")
FOR: Token = Token(TokenKind.KEYWORD, "FOR")
FOREIGN: Token = Token(TokenKind.KEYWORD, "FOREIGN")
FROM: Token = Token(TokenKind.KEYWORD, "FROM")
FULL: Token = Token(TokenKind.KEYWORD, "FULL")
GENERATED: Token = Token(TokenKind.KEYWORD, "GENERATED")
GLOB: Token = Token(TokenKind.KEYWORD, "GLOB")
GROUP: Token = Token(TokenKind.KEYWORD, "GROUP")
GROUPS: Token = Token(TokenKind.KEYWORD, "GROUPS")
HAVING: Token = Token(TokenKind.KEYWORD, "HAVING")
IF: Token = Token(TokenKind.KEYWORD, "IF")
IGNORE: Token = Token(TokenKind.KEYWORD, "IGNORE")
IMMEDIATE: Token = Token(TokenKind.KEYWORD, "IMMEDIATE")
IN: Token = Token(TokenKind.KEYWORD, "IN")
INDEX: Token = Token(TokenKind.KEYWORD, "INDEX")
INDEXED: Token = Token(TokenKind.KEYWORD, "INDEXED")
INITIALLY: Token = Token(TokenKind.KEYWORD, "INITIALLY")
INNER: Token = Token(TokenKind.KEYWORD, "INNER")
INSERT: Token = Token(TokenKind.KEYWORD, "INSERT")
INSTEAD: Token = Token(TokenKind.KEYWORD, "INSTEAD")
INTERSECT: Token = Token(TokenKind.KEYWORD, "INTERSECT")
INTO: Token = Token(TokenKind.KEYWORD, "INTO")
IS: Token = Token(TokenKind.KEYWORD, "IS")
ISNULL: Token = Token(TokenKind.KEYWORD, "ISNULL")
JOIN: Token = Token(TokenKind.KEYWORD, "JOIN")
KEY: Token = Token(TokenKind.KEYWORD, "KEY")
LAST: Token = Token(TokenKind.KEYWORD, "LAST")
LEFT: Token = Token(TokenKind.KEYWORD, "LEFT")
LIKE: Token = Token(TokenKind.KEYWORD, "LIKE")
LIMIT: Token = Token(TokenKind.KEYWORD, "LIMIT")
MATCH: Token = Token(TokenKind.KEYWORD, "MATCH")
MATERIALIZED: Token = Token(TokenKind.KEYWORD, "MATERIALIZED")
NATURAL: Token = Token(TokenKind.KEYWORD, "NATURAL")
NO: Token = Token(TokenKind.KEYWORD, "NO")
NOT: Token = Token(TokenKind.KEYWORD, "NOT")
NOTHING: Token = Token(TokenKind.KEYWORD, "NOTHING")
NOTNULL: Token = Token(TokenKind.KEYWORD, "NOTNULL")
NULL: Token = Token(TokenKind.KEYWORD_LITERAL, "NULL")
NULLS: Token = Token(TokenKind.KEYWORD, "NULLS")
OF: Token = Token(TokenKind.KEYWORD, "OF")
OFFSET: Token = Token(TokenKind.KEYWORD, "OFFSET")
ON: Token = Token(TokenKind.KEYWORD, "ON")
OR: Token = Token(TokenKind.KEYWORD, "OR")
ORDER: Token = Token(TokenKind.KEYWORD, "ORDER")
OTHERS: Token = Token(TokenKind.KEYWORD, "OTHERS")
OUTER: Token = Token(TokenKind.KEYWORD, "OUTER")
OVER: Token = Token(TokenKind.KEYWORD, "OVER")
PARTITION: Token = Token(TokenKind.KEYWORD, "PARTITION")
PLAN: Token = Token(TokenKind.KEYWORD, "PLAN")
PRAGMA: Token = Token(TokenKind.KEYWORD, "PRAGMA")
PRECEDING: Token = Token(TokenKind.KEYWORD, "PRECEDING")
PRIMARY: Token = Token(TokenKind.KEYWORD, "PRIMARY")
QUERY: Token = Token(TokenKind.KEYWORD, "QUERY")
RAISE: Token = Token(TokenKind.KEYWORD, "RAISE")
RANGE: Token = Token(TokenKind.KEYWORD, "RANGE")
RECURSIVE: Token = Token(TokenKind.KEYWORD, "RECURSIVE")
REFERENCES: Token = Token(TokenKind.KEYWORD, "REFERENCES")
REGEXP: Token = Token(TokenKind.KEYWORD, "REGEXP")
REINDEX: Token = Token(TokenKind.KEYWORD, "REINDEX")
RELEASE: Token = Token(TokenKind.KEYWORD, "RELEASE")
RENAME: Token = Token(TokenKind.KEYWORD, "RENAME")
REPLACE: Token = Token(TokenKind.KEYWORD, "REPLACE")
RESTRICT: Token = Token(TokenKind.KEYWORD, "RESTRICT")
RETURNING: Token = Token(TokenKind.KEYWORD, "RETURNING")
RIGHT: Token = Token(TokenKind.KEYWORD, "RIGHT")
ROLLBACK: Token = Token(TokenKind.KEYWORD, "ROLLBACK")
ROW: Token = Token(TokenKind.KEYWORD, "ROW")
ROWS: Token = Token(TokenKind.KEYWORD, "ROWS")
SAVEPOINT: Token = Token(TokenKind.KEYWORD, "SAVEPOINT")
SELECT: Token = Token(TokenKind.KEYWORD, "SELECT")
SET: Token = Token(TokenKind.KEYWORD, "SET")
TABLE: Token = Token(TokenKind.KEYWORD, "TABLE")
TEMP: Token = Token(TokenKind.KEYWORD, "TEMP")
TEMPORARY: Token = Token(TokenKind.KEYWORD, "TEMPORARY")
THEN: Token = Token(TokenKind.KEYWORD, "THEN")
TIES: Token = Token(TokenKind.KEYWORD, "TIES")
TO: Token = Token(TokenKind.KEYWORD, "TO")
TRANSACTION: Token = Token(TokenKind.KEYWORD, "TRANSACTION")
TRIGGER: Token = Token(TokenKind.KEYWORD, "TRIGGER")
UNBOUNDED: Token = Token(TokenKind.KEYWORD, "UNBOUNDED")
UNION: Token = Token(TokenKind.KEYWORD, "UNION")
UNIQUE: Token = Token(TokenKind.KEYWORD, "UNIQUE")
UPDATE: Token = Token(TokenKind.KEYWORD, "UPDATE")
USING: Token = Token(TokenKind.KEYWORD, "USING")
VACUUM: Token = Token(TokenKind.KEYWORD, "VACUUM")
VALUES: Token = Token(TokenKind.KEYWORD, "VALUES")
VIEW: Token = Token(TokenKind.KEYWORD, "VIEW")
VIRTUAL: Token = Token(TokenKind.KEYWORD, "VIRTUAL")
WHEN: Token = Token(TokenKind.KEYWORD, "WHEN")
WHERE: Token = Token(TokenKind.KEYWORD, "WHERE")
WINDOW: Token = Token(TokenKind.KEYWORD, "WINDOW")
WITH: Token = Token(TokenKind.KEYWORD, "WITH")
WITHOUT: Token = Token(TokenKind.KEYWORD, "WITHOUT")

INTERNED: dict[str, Token] = {
    # All operators with two character corresponds to the concatenation of
    # two single-char operators, except operators starting with !
    # To simplify lookup we bind ! with a interned token.
    "!": Token(TokenKind.UNKNOWN, "!"),
    "#": Token(TokenKind.UNKNOWN, "#"),
    # Spaces
    "\n": NEWLINE,
    "\r": Token(TokenKind.WHITESPACE, "\r"),
    " ": Token(TokenKind.WHITESPACE, " "),
    "\t": Token(TokenKind.WHITESPACE, "\t"),
    "\v": Token(TokenKind.WHITESPACE, "\v"),
    "\f": Token(TokenKind.WHITESPACE, "\f"),
    # Binary operators
    "~": BIN_NEG,
    "&": BIN_AND,
    "|": BIN_OR,
    "<<": BIN_LSHIFT,
    ">>": BIN_RSHIFT,
    # Numeric operators
    "+": NUM_PLUS,
    "-": NUM_MINUS,
    "/": NUM_DIV,
    "*": WILDCARD,
    "%": NUM_MOD,
    "^": NUM_EXP,
    # String operators
    "||": STR_CONCAT,
    # Comparison operators
    "=": CMP_EQ,
    "==": CMP_EQ2,
    "<>": CMP_NEQ,
    "<": CMP_LT,
    "<=": CMP_LEQ,
    ">": CMP_GT,
    ">=": CMP_GEQ,
    "!=": CMP_NEQ2,
    "!<": CMP_LEQ2,
    "!>": CMP_GEQ2,
    # Punctuation
    ".": DOT,
    ",": COMMA,
    ":": COLON,
    "::": DOUBLE_COLON,
    ";": SEMICOLON,
    "(": L_PAREN,
    ")": R_PAREN,
    # Known raw identifiers
    "ROWID": ROWID,
    "STRICT": STRICT,
    "FALSE": FALSE,
    "TRUE": TRUE,
    "IDENTITY": IDENTITY,
    # Keywords
    "ABORT": ABORT,
    "ACTION": ACTION,
    "ADD": ADD,
    "AFTER": AFTER,
    "ALL": ALL,
    "ALTER": ALTER,
    "ALWAYS": ALWAYS,
    "ANALYZE": ANALYZE,
    "AND": AND,
    "AS": AS,
    "ASC": ASC,
    "ATTACH": ATTACH,
    "AUTOINCREMENT": AUTOINCREMENT,
    "AUTO_INCREMENT": AUTO_INCREMENT,
    "BEFORE": BEFORE,
    "BEGIN": BEGIN,
    "BETWEEN": BETWEEN,
    "BY": BY,
    "CASCADE": CASCADE,
    "CASE": CASE,
    "CAST": CAST,
    "CHECK": CHECK,
    "COLLATE": COLLATE,
    "COLUMN": COLUMN,
    "COMMIT": COMMIT,
    "CONFLICT": CONFLICT,
    "CONSTRAINT": CONSTRAINT,
    "CREATE": CREATE,
    "CROSS": CROSS,
    "CURRENT": CURRENT,
    "CURRENT_DATE": CURRENT_DATE,
    "CURRENT_TIME": CURRENT_TIME,
    "CURRENT_TIMESTAMP": CURRENT_TIMESTAMP,
    "DATABASE": DATABASE,
    "DEFAULT": DEFAULT,
    "DEFERRABLE": DEFERRABLE,
    "DEFERRED": DEFERRED,
    "DELETE": DELETE,
    "DESC": DESC,
    "DETACH": DETACH,
    "DISTINCT": DISTINCT,
    "DO": DO,
    "DROP": DROP,
    "EACH": EACH,
    "ELSE": ELSE,
    "END": END,
    "ESCAPE": ESCAPE,
    "EXCEPT": EXCEPT,
    "EXCLUDE": EXCLUDE,
    "EXCLUSIVE": EXCLUSIVE,
    "EXISTS": EXISTS,
    "EXPLAIN": EXPLAIN,
    "FAIL": FAIL,
    "FILTER": FILTER,
    "FIRST": FIRST,
    "FOLLOWING": FOLLOWING,
    "FOR": FOR,
    "FOREIGN": FOREIGN,
    "FROM": FROM,
    "FULL": FULL,
    "GENERATED": GENERATED,
    "GLOB": GLOB,
    "GROUP": GROUP,
    "GROUPS": GROUPS,
    "HAVING": HAVING,
    "IF": IF,
    "IGNORE": IGNORE,
    "IMMEDIATE": IMMEDIATE,
    "IN": IN,
    "INDEX": INDEX,
    "INDEXED": INDEXED,
    "INITIALLY": INITIALLY,
    "INNER": INNER,
    "INSERT": INSERT,
    "INSTEAD": INSTEAD,
    "INTERSECT": INTERSECT,
    "INTO": INTO,
    "IS": IS,
    "ISNULL": ISNULL,
    "JOIN": JOIN,
    "KEY": KEY,
    "LAST": LAST,
    "LEFT": LEFT,
    "LIKE": LIKE,
    "LIMIT": LIMIT,
    "MATCH": MATCH,
    "MATERIALIZED": MATERIALIZED,
    "NATURAL": NATURAL,
    "NO": NO,
    "NOT": NOT,
    "NOTHING": NOTHING,
    "NOTNULL": NOTNULL,
    "NULL": NULL,
    "NULLS": NULLS,
    "OF": OF,
    "OFFSET": OFFSET,
    "ON": ON,
    "OR": OR,
    "ORDER": ORDER,
    "OTHERS": OTHERS,
    "OUTER": OUTER,
    "OVER": OVER,
    "PARTITION": PARTITION,
    "PLAN": PLAN,
    "PRAGMA": PRAGMA,
    "PRECEDING": PRECEDING,
    "PRIMARY": PRIMARY,
    "QUERY": QUERY,
    "RAISE": RAISE,
    "RANGE": RANGE,
    "RECURSIVE": RECURSIVE,
    "REFERENCES": REFERENCES,
    "REGEXP": REGEXP,
    "REINDEX": REINDEX,
    "RELEASE": RELEASE,
    "RENAME": RENAME,
    "REPLACE": REPLACE,
    "RESTRICT": RESTRICT,
    "RETURNING": RETURNING,
    "RIGHT": RIGHT,
    "ROLLBACK": ROLLBACK,
    "ROW": ROW,
    "ROWS": ROWS,
    "SAVEPOINT": SAVEPOINT,
    "SELECT": SELECT,
    "SET": SET,
    "TABLE": TABLE,
    "TEMP": TEMP,
    "TEMPORARY": TEMPORARY,
    "THEN": THEN,
    "TIES": TIES,
    "TO": TO,
    "TRANSACTION": TRANSACTION,
    "TRIGGER": TRIGGER,
    "UNBOUNDED": UNBOUNDED,
    "UNION": UNION,
    "UNIQUE": UNIQUE,
    "UPDATE": UPDATE,
    "USING": USING,
    "VACUUM": VACUUM,
    "VALUES": VALUES,
    "VIEW": VIEW,
    "VIRTUAL": VIRTUAL,
    "WHEN": WHEN,
    "WHERE": WHERE,
    "WINDOW": WINDOW,
    "WITH": WITH,
    "WITHOUT": WITHOUT,
}



from dataclasses import dataclass
from enum import Enum, auto
from collections.abc import Sequence
from sqlschm import tok
from typing import Iterable
import itertools


class _ReprEnum(Enum):
    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


# SQLite specific
class Dialect(_ReprEnum):
    SQLITE = auto()


class ConstraintEnforcementTime(_ReprEnum):
    DEFERRED = auto()
    IMMEDIATE = auto()


class Match(_ReprEnum):
    FULL = auto()
    PARTIAL = auto()
    SIMPLE = auto()


MATCH: frozenset[str] = frozenset(member.name for member in Match)


# SQLite specific
class OnConflict(_ReprEnum):
    ABORT = auto()
    FAIL = auto()
    IGNORE = auto()
    REPLACE = auto()
    ROLLBACK = auto()


ON_CONFLICT: frozenset[str] = frozenset(member.name for member in OnConflict)


class OnUpdateDelete(_ReprEnum):
    CASCADE = auto()
    NO_ACTION = auto()
    RESTRICT = auto()
    SET_DEFAULT = auto()
    SET_NULL = auto()


class Sorting(_ReprEnum):
    ASC = auto()
    DESC = auto()


class GeneratedKind(_ReprEnum):
    STORED = auto()
    VIRTUAL = auto()


GENERATED_KIND: frozenset[str] = frozenset(member.name for member in GeneratedKind)


"""
From the most specific to the least specific.
e.g. database.schema.table is represented as ("table", "schema", "database")
"""
QualifiedName = Sequence[str]


@dataclass(frozen=True, kw_only=True, slots=True)
class Alias:
    name: QualifiedName


@dataclass(frozen=True, kw_only=True, slots=True)
class Type:
    name: str
    params: Sequence[int] = tuple()


@dataclass(frozen=True, kw_only=True, slots=True)
class ConstraintEnforcement:
    initially: ConstraintEnforcementTime | None = None
    not_deferrable: bool = False


@dataclass(frozen=True, kw_only=True, slots=True)
class Collation:
    name: str | None = None
    value: str


@dataclass(frozen=True, kw_only=True, slots=True)
class Default:
    name: str | None = None
    expr: Sequence[tok.Token]


@dataclass(frozen=True, kw_only=True, slots=True)
class NotNull:
    name: str | None = None
    on_conflict: OnConflict | None = None


@dataclass(frozen=True, kw_only=True, slots=True)
class Generated:
    name: str | None = None
    expr: Sequence[tok.Token]
    kind: GeneratedKind | None = None


@dataclass(frozen=True, kw_only=True, slots=True)
class Indexed:
    column: str
    collation: Collation | None = None
    sorting: Sorting | None = None


@dataclass(frozen=True, kw_only=True, slots=True)
class Uniqueness:
    name: str | None = None
    is_table_constraint: bool = False
    indexed: Sequence[Indexed]
    is_primary: bool = False
    autoincrement: bool = False
    on_conflict: OnConflict | None = None


@dataclass(frozen=True, kw_only=True, slots=True)
class ForeignKey:
    name: str | None = None
    is_table_constraint: bool = False
    columns: Sequence[str]
    foreign_table: Alias
    referred_columns: Sequence[str] | None
    on_delete: OnUpdateDelete | None = None
    on_update: OnUpdateDelete | None = None
    match: Match | None = None
    enforcement: ConstraintEnforcement | None = None


@dataclass(frozen=True, kw_only=True, slots=True)
class Check:
    name: str | None = None
    is_table_constraint: bool = False
    expr: Sequence[tok.Token]


TableConstraint = Uniqueness | ForeignKey | Check


ColumnConstraint = TableConstraint | Collation | Default | NotNull | Generated


@dataclass(frozen=True, kw_only=True, slots=True)
class Column:
    name: str
    type: Type
    constraints: Sequence[ColumnConstraint]

    def collation(self, /) -> Collation | None:
        return next((x for x in self.constraints if isinstance(x, Collation)), None)

    def default(self, /) -> Default | None:
        return next((x for x in self.constraints if isinstance(x, Default)), None)

    def not_null(self, /) -> NotNull | None:
        return next((x for x in self.constraints if isinstance(x, NotNull)), None)

    def table_constraints(self, /) -> Iterable[TableConstraint]:
        return (
            x
            for x in self.constraints
            if isinstance(x, Uniqueness)
            or isinstance(x, ForeignKey)
            or isinstance(x, Check)
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class TableOptions:
    strict: bool = False
    without_rowid: bool = False


@dataclass(frozen=True, kw_only=True, slots=True)
class Table:
    name: QualifiedName
    columns: Sequence[Column]
    constraints: Sequence[TableConstraint]
    options: TableOptions = TableOptions()
    if_not_exists: bool = False
    or_replace: bool = False
    temporary: bool = False

    def all_constraints(self, /) -> Iterable[TableConstraint]:
        return itertools.chain(
            itertools.chain.from_iterable(
                col.table_constraints() for col in self.columns
            ),
            self.constraints,
        )

    def primary_key(self, /) -> Uniqueness | None:
        """First found primary key constraint"""
        return next(
            (
                x
                for x in self.all_constraints()
                if isinstance(x, Uniqueness) and x.is_primary
            ),
            None,
        )

    def uniqueness(self, /) -> list[Uniqueness]:
        """All uniqueness constraints, including the primary key"""
        return [x for x in self.all_constraints() if isinstance(x, Uniqueness)]

    def foreign_keys(self, /) -> list[ForeignKey]:
        """All foreign key constraints"""
        return [x for x in self.all_constraints() if isinstance(x, ForeignKey)]

    def checks(self, /) -> list[Check]:
        """All foreign key constraints"""
        return [x for x in self.all_constraints() if isinstance(x, Check)]


@dataclass(frozen=True, kw_only=True, slots=True)
class Schema:
    tables: Sequence[Table]


Symbols = dict[str, Table]


def symbols(schema: Schema, /) -> Symbols:
    return dict((tbl.name[0], tbl) for tbl in schema.tables)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqlschm']

package_data = \
{'': ['*']}

modules = \
['py']
entry_points = \
{'console_scripts': ['corpus = scripts:main']}

setup_kwargs = {
    'name': 'sqlschm',
    'version': '0.5.0',
    'description': 'A SQLite Schema parser',
    'long_description': None,
    'author': 'Victorien Elvinger',
    'author_email': 'victorien.elvinger@inria.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/coast-team/sqlschm',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
